<?php

namespace Drupal\Tests\xero\Kernel\Normalizer;

use Drupal\Core\TypedData\TypedDataTrait;
use Drupal\KernelTests\KernelTestBase;
use Drupal\serialization\Normalizer\ComplexDataNormalizer;
use Drupal\serialization\Normalizer\TypedDataNormalizer;
use Drupal\xero\Normalizer\XeroListNormalizer;
use Drupal\xero\Normalizer\XeroNormalizer;
use Symfony\Component\Serializer\Serializer;

/**
 * Tests denormalizer with services enabled.
 *
 * @group xero
 */
class XeroNestedNormalizerTest extends KernelTestBase {

  use TypedDataTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['serialization', 'xero'];

  /**
   * Asserts that denormalization works for deeply-nested objects.
   */
  public function testDenormalize() {
    $data = [
      'TrackingCategories' => [
        'TrackingCategory' => [
          [
            'Name' => 'Region',
            'Status' => 'ACTIVE',
            'TrackingCategoryID' => '351953c4-8127-4009-88c3-f9cd8c9cbe9f',
            'Options' => [
              'Option' => [
                ['Name' => 'West Coast', 'Status' => 'ACTIVE'],
                ['Name' => 'Eastside', 'Status' => 'ACTIVE'],
              ],
            ],
          ],
        ],
      ],
    ];
    $normalizer = new XeroNormalizer($this->getTypedDataManager());
    $denormalizedData = $normalizer->denormalize(
      $data,
      '\Drupal\xero\Plugin\DataType\TrackingCategory',
      'xml',
      ['plugin_id' => 'xero_tracking']
    );

    $values = $denormalizedData->getValue();
    $this->assertArrayHasKey('Options', $values[0]);
    $this->assertArrayNotHasKey('Option', $values[0]['Options']);
    $this->assertCount(2, $denormalizedData->get(0)->get('Options'));
  }

  /**
   * Tests normalization of Xero contact data under different data scenarios.
   *
   * @param array $expected
   *   The expected normalized value.
   * @param array $data
   *   The data for the typed data object.
   * @param string $message
   *   An optional message to display for the test.
   *
   * @dataProvider contactProvider
   */
  public function testNormalizeContact(array $expected, array $data, $message) {
    $this->assertNormalizationSuccessful(
      $expected, $message, 'xero_contact', 'Contact', $data, []
    );
    $this->assertNormalizationSuccessful(
      $expected, $message, 'xero_contact', 'Contact', [], $data
    );
  }

  /**
   * Tests normalization of contact when it is updated with partial data.
   */
  public function testNormalizePartialContact() {
    $definition = $this->getTypedDataManager()
      ->createDataDefinition('xero_contact');
    $normalizer = $this->getNormalizer();

    $initial = [
      'Name' => 'ABC Limited',
      'Phones' => [
        ['PhoneType' => 'DEFAULT'],
        ['PhoneType' => 'MOBILE'],
      ],
    ];

    $item = $this->getTypedDataManager()
      ->create($definition, $initial, 'Contact');
    $phoneItem = $item->get('Phones')->get(0);
    $phoneItem->get('PhoneNumber')->setValue('1111111');
    $phoneItem->get('PhoneAreaCode')->setValue('04');
    $phoneItem->get('PhoneCountryCode')->setValue('64');

    $normalized = $normalizer->normalize($item, 'json', [
      'plugin_id' => 'xero_contact',
    ]);
    $expected = [
      'Contact' => [
        'Name' => 'ABC Limited',
        'Phones' => [
          [
            'PhoneType' => 'DEFAULT',
            'PhoneNumber' => '1111111',
            'PhoneAreaCode' => '04',
            'PhoneCountryCode' => '64',
          ],
          [
            'PhoneType' => 'MOBILE',
          ],
        ],
      ],
    ];

    $this->assertSame($expected, $normalized);
  }

  /**
   * Tests normalization of Xero invoice data under different data scenarios.
   *
   * @param array $expected
   *   The expected normalized value.
   * @param array $data
   *   The data to for the typed data object.
   * @param string $message
   *   An optional message to display for the test.
   *
   * @dataProvider invoiceProvider
   */
  public function testNormalizeInvoice(array $expected, array $data, $message) {
    $this->assertNormalizationSuccessful(
      $expected, $message, 'xero_invoice', 'Invoice', $data, []
    );
    $this->assertNormalizationSuccessful(
      $expected, $message, 'xero_invoice', 'Invoice', [], $data
    );
  }

  /**
   * Provides test arguments for testNormalizeContact.
   *
   * @return array
   *   An indexed array of tests to run with test arguments.
   */
  public function contactProvider() {
    return [
      [
        [],
        [],
        'empty data is empty',
      ],
      [
        ['Contact' => ['Name' => 'ABC Limited']],
        ['Name' => 'ABC Limited'],
        'single properties property are normalized with value.',
      ],
      [
        [
          'Contact' => [
            'Name' => 'ABC Limited',
            'FirstName' => 'Ben',
            'LastName' => 'Bowden',
          ],
        ],
        [
          'Name' => 'ABC Limited',
          'FirstName' => 'Ben',
          'LastName' => 'Bowden',
        ],
        'multiple and optional properties are normalized with values.',
      ],
      [
        [
          'Contact' => [
            'FirstName' => 'Ben',
            'LastName' => 'Bowden',
          ],
        ],
        [
          'FirstName' => 'Ben',
          'LastName' => 'Bowden',
        ],
        'required properties are excluded if unspecified.',
      ],
      [
        [
          'Contact' => [
            'Name' => 'ABC Limited',
            'FirstName' => NULL,
            'LastName' => '',
          ],
        ],
        [
          'Name' => 'ABC Limited',
          'FirstName' => NULL,
          'LastName' => '',
        ],
        'empty properties are normalized as empty values.',
      ],
      [
        [
          'Contact' => [
            'Name' => 'ABC Limited',
            'Phones' => [
              [
                'PhoneAreaCode' => '083',
                'PhoneNumber' => '111',
              ],
            ],
          ],
        ],
        [
          'Name' => 'ABC Limited',
          'Phones' => [
            [
              'PhoneAreaCode' => '083',
              'PhoneNumber' => '111',
            ],
          ],
        ],
        'a specified nested xero property is normalized with nested values',
      ],
      [
        [
          'Contact' => [
            'Name' => 'ABC Limited',
            'Phones' => [],
          ],
        ],
        [
          'Name' => 'ABC Limited',
          'Phones' => [],
        ],
        'explicitly empty nested properties are normalized as empty arrays',
      ],
      [
        [
          'Contact' => [
            'Name' => 'ABC Limited',
            'Phones' => [],
          ],
        ],
        [
          'Name' => 'ABC Limited',
          'Phones' => NULL,
        ],
        'explicitly nulled nested properties are normalized as empty arrays',
      ],
    ];
  }

  /**
   * Provides test arguments for testNormalizeInvoice.
   *
   * @return array
   *   An indexed array of tests to run with test arguments.
   */
  public function invoiceProvider() {
    return [
      [
        [],
        [],
        'empty data is empty',
      ],
      [
        ['Invoice' => ['Type' => 'ACCREC']],
        ['Type' => 'ACCREC'],
        'single properties property are normalized with value.',
      ],
      [
        [
          'Invoice' => [
            'Type' => 'ACCPAY',
            'Date' => '2011-10-12',
            'Reference' => 'Inv23123',
          ],
        ],
        [
          'Type' => 'ACCPAY',
          'Date' => '2011-10-12',
          'Reference' => 'Inv23123',
        ],
        'multiple and optional properties are normalized with values.',
      ],
      [
        [
          'Invoice' => [
            'Type' => 'ACCPAY',
            'InvoiceNumber' => NULL,
            'Reference' => '',
          ],
        ],
        [
          'Type' => 'ACCPAY',
          'InvoiceNumber' => NULL,
          'Reference' => '',
        ],
        'empty properties are normalized as empty values.',
      ],
      [
        [
          'Invoice' => [
            'Type' => 'ACCPAY',
            'Contact' => [
              'Name' => 'ABC Limited',
              'FirstName' => 'Ben',
            ],
          ],
        ],
        [
          'Type' => 'ACCPAY',
          'Contact' => [
            'Name' => 'ABC Limited',
            'FirstName' => 'Ben',
          ],
        ],
        'a specified nested xero property is normalized with nested values',
      ],
      [
        [
          'Invoice' => [
            'Type' => 'ACCPAY',
            'LineItems' => [],
            'Payments' => [],
          ],
        ],
        [
          'Type' => 'ACCPAY',
          'LineItems' => [],
          'Payments' => NULL,
        ],
        'explicitly empty nested list properties are normalized as empty arrays',
      ],
    ];
  }

  /**
   * Test setting to null the contact on the invoice.
   *
   * This property has the unusual characteristic that its normalized output
   * varies depending on how it was set to null.
   */
  public function testInvoiceContact() {
    $data = [
      'Type' => 'ACCPAY',
      'Contact' => NULL,
    ];

    $message = 'nested property explicitly set to null is normalized as null';
    $expected = [
      'Invoice' => [
        'Type' => 'ACCPAY',
        'Contact' => NULL,
      ],
    ];
    $this->assertNormalizationSuccessful(
      $expected, $message, 'xero_invoice', 'Invoice', [], $data
    );

    $message = 'nested property explicitly created as null is normalized to empty array';
    $expected = [
      'Invoice' => [
        'Type' => 'ACCPAY',
        'Contact' => [],
      ],
    ];
    $this->assertNormalizationSuccessful(
      $expected, $message, 'xero_invoice', 'Invoice', $data, []
    );
  }

  /**
   * Tests the normalization of a Xero type under different data scenarios.
   *
   * @param array $expected
   *   The expected normalized value.
   * @param string $message
   *   An message to display for the test.
   * @param string $xeroType
   *   The Xero type to test.
   * @param string $xeroTypeLabel
   *   The label of the Xero type to test.
   * @param array $initialData
   *   (optional) The data to create the typed data object with.
   * @param array $subsequentData
   *   (optional) The data to set subsequently on the typed data object.
   */
  public function assertNormalizationSuccessful(array $expected, $message, $xeroType, $xeroTypeLabel, array $initialData = [], array $subsequentData = []) {
    $definition = $this->getTypedDataManager()
      ->createDataDefinition($xeroType);

    $normalizer = $this->getNormalizer();

    $item = $this->getTypedDataManager()->create($definition, $initialData, $xeroTypeLabel);
    foreach ($subsequentData as $name => $value) {
      $item->get($name)->setValue($value);
    }

    $normalized = $normalizer->normalize($item, 'json', [
      'plugin_id' => $xeroType,
    ]);
    // Use assertSame as data type may matter.
    $this->assertSame($expected, $normalized, $message);
  }

  /**
   * Tests serialization of Xero contact data under different data scenarios.
   *
   * @param string $jsonExpected
   *   The expected serialized json.
   * @param string $xmlExpected
   *   The expected serialized xml.
   * @param array $data
   *   The data for the typed data object to be serialized.
   * @param string $message
   *   An optional message to display for the test.
   *
   * @dataProvider serializationProvider
   */
  public function testSerialization($jsonExpected, $xmlExpected, array $data, $message) {
    $serializer = \Drupal::service('serializer');
    $definition = $this->getTypedDataManager()
      ->createDataDefinition('xero_contact');
    $item = $this->getTypedDataManager()->create($definition, $data, 'Contact');
    $context = ['plugin_id' => 'xero_contact', 'xml_root_node_name' => 'Contacts'];
    $json = $serializer->serialize($item, 'json', $context);
    $this->assertEquals($jsonExpected, $json, $message . " (json)");
    $xml = $serializer->serialize($item, 'xml', $context);
    $this->assertEquals($xmlExpected, $xml, $message . " (xml)");
  }

  /**
   * Provides test arguments for testSerialization.
   *
   * @return array
   *   An indexed array of tests to run with test arguments.
   */
  public function serializationProvider() {
    return [
      [
        '{"Contact":{"Name":"ABC Limited","FirstName":"","LastName":null,"Phones":[]}}',
        "<?xml version=\"1.0\"?>\n<Contacts><Contact><Name>ABC Limited</Name><FirstName></FirstName><LastName/><Phones/></Contact></Contacts>\n",
        [
          'Name' => 'ABC Limited',
          'FirstName' => '',
          'LastName' => NULL,
          'Phones' => [],
        ],
        'serializer uses normalizer and handles empty values',
      ],
    ];
  }

  /**
   * Gets the normalizer.
   *
   * @return \Drupal\xero\Normalizer\XeroNormalizer
   *   The normalizer instance.
   */
  protected function getNormalizer() {
    $normalizer = new XeroNormalizer($this->typedDataManager);
    $listNormalizer = new XeroListNormalizer();
    $typeddata_normalizer = new TypedDataNormalizer();
    $complex_normalizer = new ComplexDataNormalizer();
    $serializer = new Serializer([
      $listNormalizer,
      $normalizer,
      $typeddata_normalizer,
      $complex_normalizer,
    ]);
    $normalizer->setSerializer($serializer);

    return $normalizer;
  }

}
